/*
  CaboCha -- Yet Another Japanese Dependency Structure Analyzer

  $Id: dep.cpp,v 1.14 2005/12/24 12:30:35 taku Exp $;

  Copyright (C) 2000-2004 Taku Kudo <taku-ku@is.aist-nara.ac.jp>
  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "dep.h"
#include "param.h"
#include "common.h"
#include <iterator>

#define MAX_GAP_SIZE ((size_t)7)

namespace CaboCha
{
  bool DepAnalyzer::open (Param &p)
  {
    action_mode     = p.getProfileInt ("action-mode");
    partial_parsing = p.getProfileInt ("candidate");
    std::string filename = p.getProfileString ("parser-model");

    if (action_mode == PARSER_MODE) {

      if (filename.empty()) {
  _what = "DepAnalyzer::open (): model file is empty";
  return false;
      }

      model = yamcha_svm_new ((char *)filename.c_str());
      if (! model) {
  _what = "DepAnalyzer::DepAnalyzer(): ";
  _what += yamcha_svm_strerror (model);
  return false;
      }

      csize = yamcha_svm_get_class_size (model);
    }

    return true;
  }

  bool DepAnalyzer::estimate (int src, int dst, std::string &rel, double &score)
  {
    if (action_mode == PARSER_MODE && result[src].link != -1) {
      if (chunks[src]->link == dst) {
  rel = chunks[src]->rel;
  return true;
      } else {
  rel = "O";
  return false;
      }
    }

    fpset.clear();
    int dist = dst - src;
    if (dist == 1)                   fpset.insert ("DIST:1");
    else if (dist >= 2 && dist <= 5) fpset.insert ("DIST:2-5");
    else                             fpset.insert ("DIST:6-");

    // static feature
    for (size_t i = 0; i < static_feature[src].size(); i++) {
      static_feature[src][i][0] = 'f';
      fpset.insert (static_feature[src][i]);
    }

    for (size_t i = 0; i < static_feature[dst].size(); i++) {
      static_feature[dst][i][0] = 'F';
      fpset.insert (static_feature[dst][i]);
    }

    size_t max_gap = 0;

    // gap feature
    int k = dst * (dst + 1) / 2 + src;
    max_gap = _min(gap[k].size(), MAX_GAP_SIZE);
    for (size_t i = 0; i < max_gap; ++i)
      fpset.insert (gap[k][i]);

    // dynamic features
    max_gap = _min(dyn_a[dst].size(), MAX_GAP_SIZE);
    for (size_t i = 0; i < max_gap; ++i) {
      dyn_a[dst][i][0] = 'A';
      fpset.insert (dyn_a[dst][i]);
    }

    max_gap = _min(dyn_a[src].size(), MAX_GAP_SIZE);
    for (size_t i = 0; i < max_gap; ++i) {
      dyn_a[src][i][0] = 'a';
      fpset.insert (dyn_a[src][i]);
    }

    max_gap = _min(dyn_b[dst].size(), MAX_GAP_SIZE);
    for (size_t i = 0; i < max_gap; ++i) {
      dyn_b[dst][i][0] = 'B';
      fpset.insert (dyn_b[dst][i]);
    }

    size_t fsize = fpset.size();
    std::copy (fpset.begin(), fpset.end(), fp);

    // sort + uniq
    if (action_mode == PARSER_MODE) {

      if (partial_parsing) {
  int d = chunks[src]->link;
  if (d != -1) {
    score = 0.0;
    rel = chunks[src]->rel;
    return (d == dst);
  }
      }

      double max_score = 0.0;
      size_t id  = 0;

      yamcha_result_t *r = yamcha_svm_classify (model, fsize, fp);

      for (size_t i = 0; i < csize; i++) {
  if (max_score < r[i].score) { // get max item only
    max_score = r[i].score;
    id = i; // save current id
  }
      }

      score = r[id].dist;
      rel  = r[id].name;

      return (rel != "O");

    } else {

      bool isdep = false;
      rel = "O";

      if (chunks[src]->link == dst) {
  isdep = true;
  rel   = chunks[src]->rel;
  if (rel.empty ()) throw std::runtime_error ("DepAnalyzer::estimate (): empty relation");
      }

      if (! is_write_header) {
  std::cout << "Version: " << VERSION << std::endl;
  std::cout << "Package: " << PACKAGE << std::endl;
  std::cout << std::endl;
  is_write_header = 1;
      }

      std::cout << rel << " ";
      std::copy (fp, fp + fsize, std::ostream_iterator<char *>(std::cout, " "));
      std::cout << std::endl;

      return isdep;
    }

    return false;
  }

  void DepAnalyzer::resize (size_t size)
  {
    static_feature.clear();
    static_feature.resize(size);

    dyn_b_feature.clear();
    dyn_b_feature.resize(size);
    dyn_b.clear();
    dyn_b.resize(size);

    dyn_a_feature.clear();
    dyn_a_feature.resize(size);
    dyn_a.clear();
    dyn_a.resize(size);

    gap.clear();
    gap_list.clear();
    gap_list.resize(size);

    result.resize (size);
    std::for_each (result.begin(), result.end(), std::mem_fun_ref (&result_t::reset));
  }

  size_t DepAnalyzer::build (Tree *tree)
  {
    size_t size  = tree->size ();
    chunks.clear ();

    for (size_t i = 0; i < size; i++)
      if ((*tree)[i].chunk) chunks.push_back ((*tree)[i].chunk);

    if (chunks.empty()) return 0;

    resize (chunks.size());

    // make static and dynamic features;
    size_t ci = 0;
    for (std::vector<Chunk *>::iterator it = chunks.begin(); it != chunks.end(); ++it) {
      for (char **n = (*it)->feature; *n; ++n) {
  char * feature = *n;
  switch (feature[0]) {
  case 'F': static_feature[ci].push_back(feature); break;
  case 'G': gap_list[ci].push_back(feature);       break;
  case 'A': dyn_a_feature[ci].push_back(feature);  break;
  case 'B': dyn_b_feature[ci].push_back(feature);  break;
  default: throw std::runtime_error (std::string ("DepAnalyzer::build():  unkown feature type") + feature);
  }
      }
      ++ci;
    }

    // make gap features
    size = chunks.size ();
    gap.resize (size*(size+3)/2+1);

    for (size_t k = 0; k < size; k++)
      for (size_t i = 0; i < k; i++)
  for (size_t j = k + 1; j < size; j++)
    std::copy (gap_list[k].begin(), gap_list[k].end(), std::back_inserter(gap[j*(j+1)/2+i]));

    return size;
  }

  Tree *DepAnalyzer::parse (Tree *tree)
  {
    size_t size = build (tree);

    if (! size) return tree;

    std::vector <int> dead  (size);
    std::fill (dead.begin(),  dead.end(), 0);

    std::vector <int> alive (size);

    // main loop
    for (;;) {

      alive.clear();

      for (size_t i = 0; i < dead.size(); i++)
  if (! dead[i]) alive.push_back(i);

      if (alive.size() <= 1) break;

      bool prev   = false; // if previous chunk does NOT depend on the next true;
      bool change = false; // if current seeking does NOT change the status, true

      int lsize   = (int)(alive.size()-1); // size of alive, ignore the last chunk
      int lindex  = (int)(alive.size()-2); // last index of alive, ignore tha last chunk

      for (int i = 0; i < lsize; i++) {
  int    src    = alive[i];
  int    dst    = alive[i+1];
  bool   isdep  = true;
  double score  = 0.0;
  std::string rel  = "D";

  if (src != (int)size-2 && result[src].link != dst)
    isdep = estimate (src, dst, rel, score);

  // force to change TRUE
  if (i == lindex && ! change) isdep = true;

  // OK depend
  if (isdep) {

    // make dynamic features
    if (result[src].link != dst) {
      std::copy (dyn_b_feature[dst].begin(), dyn_b_feature[dst].end(), std::back_inserter (dyn_b[src]));
      std::copy (dyn_a_feature[src].begin(), dyn_a_feature[src].end(), std::back_inserter (dyn_a[dst]));
      result[src].score = score;
      result[src].rel = rel;
      result[src].link = dst;
    }

    if (! dead[src]) change    = true;
    if (! prev)      dead[src] = true;
  }

  prev = isdep;
      }
    }

    // write result
    for (size_t i = 0; i < size; i++) {
      chunks[i]->link  = result[i].link;
      chunks[i]->score = result[i].score;
      chunks[i]->rel   = tree->alloc (const_cast<char *>(result[i].rel.c_str()));
    }

    tree->outputLayer = OUTPUT_PARSE;

    return tree;
  }
}
