/*
  CaboCha -- Yet Another Japanese Dependency Structure Analyzer

  $Id: ne.cpp,v 1.11 2005/12/24 12:30:35 taku Exp $;

  Copyright (C) 2000-2004 Taku Kudo <taku-ku@is.aist-nara.ac.jp>
  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#if defined  (_WIN32) && ! defined (__CYGWIN__)
#include "ne-sjis.h"
#else
#include "ne.h"
#endif

#include "param.h"
#include <iterator>
#include <cstdio>

namespace CaboCha
{
  enum
    {
      OTHER,
      KANJI,
      SYMBOL,
      HYPHEN,
      MONEY,
      ALPHANUMERIC,
      KANJINUMERIC,
      CAPALPHA,
      ALPHA,
      HIRAGANA,
      KATAKANA,
      GREEK,
      CYRILLIC,
      SPACE
    };

  static int kuten_table [96][96];
  static int ascii_table [256];
  static bool isload_table = false;

  static inline void euc2kuten (unsigned int &hib, unsigned int &lob)
  {
    hib -= 160; lob -= 160;
  }

  static inline void sjis2kuten (unsigned int &hib, unsigned int &lob)
  {
    unsigned char c1 = hib;
    unsigned char c2 = lob;
    int adjust = c2 < 159;
    int rowOffset = c1 < 160 ? 112 : 176;
    int cellOffset = adjust ? (c2 > 127 ? 32 : 31) : 126;
    hib = (((c1 - rowOffset) << 1) - adjust) - 32;
    lob -= (cellOffset + 32);
  }

  static void init_table ()
  {
    // JISX 2080-1990 Kuten code table
    for (int i = 0; i <= 94; ++i)
      for (int j = 0; j <= 94; ++j)
	kuten_table[i][j] = OTHER;

    // initilize all char as KANJI
    for (int i = 16; i <= 94; ++i)
      for (int j = 1; j <= 94; ++j)
	kuten_table[i][j] = KANJI;

    kuten_table[1][1] = SPACE;
    for (int i = 2;  i <= 94; ++i) kuten_table[1][i] = SYMBOL;
    for (int i = 79; i <= 82; ++i) kuten_table[1][i] = MONEY;
    for (int i = 1;  i <= 84; ++i) kuten_table[2][i] = SYMBOL;
    for (int i = 16; i <= 25; ++i) kuten_table[3][i] = ALPHANUMERIC;
    for (int i = 33; i <= 58; ++i) kuten_table[3][i] = CAPALPHA;
    for (int i = 65; i <= 90; ++i) kuten_table[3][i] = ALPHA;
    for (int i = 1;  i <= 83; ++i) kuten_table[4][i] = HIRAGANA;
    for (int i = 1;  i <= 86; ++i) kuten_table[5][i] = KATAKANA;
    for (int i = 1;  i <= 56; ++i) kuten_table[6][i] = GREEK;
    for (int i = 1;  i <= 81; ++i) kuten_table[7][i] = CYRILLIC;

    //  (üʿ񤤤򤹤)
    kuten_table[1][28] = HYPHEN;

    // 󻰻͸ϻȬ彽ɴ
    kuten_table[16][76] = KANJINUMERIC;
    kuten_table[38][83] = KANJINUMERIC;
    kuten_table[27][16] = KANJINUMERIC;
    kuten_table[27][45] = KANJINUMERIC;
    kuten_table[24][62] = KANJINUMERIC;
    kuten_table[47][27] = KANJINUMERIC;
    kuten_table[28][23] = KANJINUMERIC;
    kuten_table[40][12] = KANJINUMERIC;
    kuten_table[22][69] = KANJINUMERIC;
    kuten_table[29][29] = KANJINUMERIC;
    kuten_table[41][20] = KANJINUMERIC;
    kuten_table[32][73] = KANJINUMERIC;
    kuten_table[43][92] = KANJINUMERIC;
    kuten_table[18][15] = KANJINUMERIC;
    kuten_table[35][91] = KANJINUMERIC;

    // ascii table
    for (int i = 0;   i <= 255;  ++i) ascii_table[i] = OTHER;
    for (int i = 33;  i <= 47;   ++i) ascii_table[i] = SYMBOL;
    for (int i = 48;  i <= 57;   ++i) ascii_table[i] = ALPHANUMERIC;
    for (int i = 58;  i <= 64;   ++i) ascii_table[i] = SYMBOL;
    for (int i = 65;  i <= 90;   ++i) ascii_table[i] = CAPALPHA;
    for (int i = 91;  i <= 96;   ++i) ascii_table[i] = SYMBOL;
    for (int i = 97;  i <= 122;  ++i) ascii_table[i] = ALPHA;
    for (int i = 123; i <= 126;  ++i) ascii_table[i] = SYMBOL;

    ascii_table[36] = MONEY; //
    ascii_table[92] = MONEY; //
  }

  static int getCharClass (char *str, char *end, char * &next)
  {
    unsigned int hb = 0;
    unsigned int lb = 0;
    unsigned int len = end - str;

#if defined  (_WIN32) && ! defined (__CYGWIN__)

    if (static_cast<unsigned char>(str[0]) <= 0x80) {
      next = str + 1;
      hb = static_cast<unsigned char>(str[0]);
      return ascii_table[hb];
    } else if (static_cast<unsigned char>(str[0]) >= 0xa0 &&
	       static_cast<unsigned char>(str[0]) <= 0xdf) {
      next = str + 1;
      return KATAKANA;
    } else if (len >= 2 && (str[0] & 0x80)) {
      next = str + 2;
      hb = static_cast<unsigned char>(str[0]);
      lb = static_cast<unsigned char>(str[1]);
      sjis2kuten (hb, lb);
      return kuten_table[hb][lb];
    } else {
      next = str + 1;
      return OTHER;
    }

#else

    if (len >= 3 && static_cast<unsigned char>(str[0]) == 0x8f && (str[1] & 0x80) && (str[2] & 0x80)) {
      next = str + 3;
      return KANJI; // 3-byte EUC, set all type as KANJI, ad-hoc
    } else if (len >= 2 && (str[0] & 0x80) && (str[1] & 0x80)) {
      next = str + 2;
      hb = static_cast<unsigned char>(str[0]);
      lb = static_cast<unsigned char>(str[1]);
      if (hb  == 142 && lb >= 166  && lb <= 222) return KATAKANA;  // EUC Hankaku Katakana
      euc2kuten (hb, lb);
      return kuten_table[hb][lb];
    } else {
      next = str + 1;
      hb = static_cast<unsigned char>(str[0]);
      return ascii_table[hb];
    }

#endif

    return OTHER;
  }

  static void getCharFeature (char *str, char *feature)
  {
    size_t len = std::strlen (str);
    char *end = str + len;
    char *next = 0;
    int type1 = -1;
    int type2 = -1;
    size_t clen = 0;
    int iscap = false;

    while (str != end) {
      int c = getCharClass (str, end, next);
      str = next;

      if (iscap && c != ALPHA) iscap = false;

      if (clen == 0) {
	type1 = c;
      } else if (clen == 1 && type1 == CAPALPHA && c == ALPHA) {
	type2 = type1;
	type1 = c;
	iscap = true;
      } else if (c == HYPHEN) {
	// ignore
      } else if (c != type1) {
	type2 = type1;
	type1 = c;
      }

      ++clen;
    }

    // 1 ʸ϶̤.
    // 2ʸʾξ:
    //      Ʊʸ
    //      ۼʸ -> Ǹ2ʸϢܤȤ

    if (iscap) {
      std::sprintf (feature, "C");
    } else if (clen == 1) {
      std::sprintf (feature, "S%d", type1);
    } else if (type2 == -1) {
      std::sprintf (feature, "M%d",  type1);
    } else {
      std::sprintf (feature, "T%d-%d", type2, type1);
    }

    return;
  }

  bool NE::open (Param &p)
  {
    partial_parsing = p.getProfileInt ("candidate");
    action_mode     = p.getProfileInt ("action-mode");
    std::string filename = p.getProfileString ("ne-model");

    if (action_mode == PARSER_MODE) {
      if (filename.empty()) {
	_what = "NE::open (): model file is empty";
	return false;
      }

      char *tmp[] = {0, "-m", 0, "-C", 0};
      tmp[0] = const_cast<char *>(p.getSystemName());
      tmp[2] = const_cast<char *>(filename.c_str());

      chunker = yamcha_new (4, tmp);
      if (! chunker) {
	_what = std::string("NE::open ():") + yamcha_strerror (chunker);
	return false;
      }
    }

    if (! isload_table) init_table ();
    isload_table = true;

    return true;
  }

  Tree *NE::parse (Tree *tree)
  {
    if (!tree) return 0;

    size_t size  = tree->size();

    yamcha_clear (chunker);

    size_t composit_flg = 0;

    for (size_t i = 0; i < size; i++) {

      feature.clear();

      if (strncmp ((*tree)[i].pos, NE_COMPOSIT, strlen((*tree)[i].pos)) == 0) ++composit_flg;
      else  composit_flg = 0;

      if (composit_flg >= 2) continue;

      const char *surface = (*tree)[i].surface;
      getCharFeature (const_cast<char *>(surface), char_feature);

      feature.push_back (surface);
      feature.push_back (char_feature);
      feature.push_back ((*tree)[i].pos);
      const char *ne = (*tree)[i].ne;

      if (action_mode == PARSER_MODE) {
	if (partial_parsing && ne[0] != '\0') feature.push_back (ne);
	yamcha_add (chunker,feature.size(), (char **)&feature[0]);

      } else {
	std::copy (feature.begin(), feature.end(), std::ostream_iterator<const char*>(std::cout, " "));
	std::cout << ne << '\n';
      }
    }

    if (action_mode == PARSER_MODE) {

      yamcha_parse (chunker);

      const char *prev = 0;
      size_t composit_flg = 0;
      size_t ci = 0;

      for (size_t i = 0; i < size; i++) {

	if (strcmp ((*tree)[i].pos, NE_COMPOSIT) == 0)  ++composit_flg;
	else  composit_flg = 0;

	if (composit_flg >= 2) {
	  (*tree)[i].ne = tree->alloc (const_cast<char *>(prev));
	  if ((*tree)[i].ne[0] != 'O') (*tree)[i].ne[0] = 'I';
	} else {
	  char *p = yamcha_get_tag (chunker, ci);
          (*tree)[i].ne = tree->alloc (p ? p : (char *)"O");
	  ++ci;
	}

	// replace invalid sequence (2003/08/08) from tetsu-ta
	if (prev && prev[0] == 'I'
	    && (*tree)[i].ne[0] == 'I'
	    && strlen ((*tree)[i].ne) >= 3
	    && strlen (prev) >= 3
	    && strcmp ((*tree)[i].ne + 3, prev + 3)) (*tree)[i].ne[0] = 'B';

	if (i == 0 && (*tree)[i].ne[0] == 'I') (*tree)[i].ne[0] = 'B';

	prev = (*tree)[i].ne;
      }

    } else {
      std::cout << '\n';
    }

    return tree;
  }
}
