/*
  CaboCha -- Yet Another Japanese Dependency Structure Analyzer

  $Id: parser.cpp,v 1.27 2005/12/24 12:30:35 taku Exp $;

  Copyright (C) 2000-2004 Taku Kudo <taku-ku@is.aist-nara.ac.jp>
  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "cabocha.h"
#include "param.h"
#include "common.h"
#if defined  (_WIN32) && ! defined (__CYGWIN__)
#include "ne-sjis.h"
#else
#include "ne.h"
#endif
#include "chunker.h"
#include "dep.h"
#include "selector.h"
#include "chasenio.h"
#include "mecabio.h"
#include <strstream>
#include <fstream>

namespace CaboCha {

  static const Option long_options[] = {
    {"action-mode",     't', 0, "TYPE", "set action mode\n\t\t\t    0 - parser mode (default)\n\t\t\t    1 - training mode"},
    {"output-format",   'f', 0, "TYPE", "set output format style\n\t\t\t    0 - tree (default)\n\t\t\t    1 - lattice\n\t\t\t    2 - tree + lattice\n\t\t\t    3 - XML"},
    {"input-layer",     'I', 0, "LAYER", "set input layer\n\t\t\t    0 - raw sentence layer (default)\n\t\t\t    1 - POS tagged layer\n\t\t\t    2 - POS tagger and Chunked layer\n\t\t\t    3 - POS tagged, Chunked and Feature selected layer"},
    {"output-layer",    'O', 0, "LAYER", "set output layer\n\t\t\t    1 - POS tagged layer\n\t\t\t    2 - POS tagged and Chunked layer\n\t\t\t    3 - POS tagged, Chunked and Feature selected layer\n\t\t\t    4 - Parsed layer (default)"},
    {"ne",              'n', 0, "MODE",  "output NE tag\n\t\t\t    0 - without NE (default)\n\t\t\t    1 - output NE with chunk constarint\n\t\t\t    2 - output NE without chunk constarint"},
    {"parser-model",    'm', 0, "FILE", "use FILE as parser model file"},
    {"chunker-model",   'M', 0, "FILE", "use FILE as chunker model file"},
    {"ne-model",        'N', 0, "FILE", "use FILE as NE tagger model file"},
    {"candidate",       'C', 0, 0,      "partial parsing mode"},
    {"rcfile",          'r', 0, "FILE", "use FILE as resource file"},
    {"chasenrc",        'c', 0, "FILE", "use FILE as chasenrc"},
    {"mecabrc",         'b', 0, "FILE", "use FILE as mecabrc"},
    {"morphological-analyzer", 'a', 0, "TYPE", "set morphological analyzer, chasen|mecab"},
    {"output",          'o', 0, "FILE", "use FILE as output file"},
    {"version",         'v', 0, 0, "show the version and exit"},
    {"help",            'h', 0, 0, "show this help and exit"},
    {0, 0, 0, 0}
  };

#if defined  (_WIN32) && ! defined (__CYGWIN__)
  static std::string getWindowsRegValue (const char *key,
					 const char *name)
  {
    HKEY hKey;
    char v[1024];
    DWORD vt;
    DWORD size = sizeof (v);

    RegOpenKeyEx (HKEY_LOCAL_MACHINE,
		  key, 0, KEY_READ, &hKey);
    RegQueryValueEx (hKey, name, 0, &vt, (BYTE *)v, &size);
    RegCloseKey (hKey);
    if (vt == REG_SZ) {
      return std::string (v);
    }

    RegOpenKeyEx (HKEY_CURRENT_USER,
		  key, 0, KEY_READ, &hKey);
    RegQueryValueEx (hKey, name, 0, &vt, (BYTE *)v, &size);
    RegCloseKey (hKey);
    if (vt == REG_SZ) {
      return std::string (v);
    }

    return std::string ("");
  }
#endif

  static std::string getDefaultRc ()
  {
#ifdef HAVE_GETENV
    char *homedir = getenv("HOME");
    if (homedir) {
      std::string s = createFileName (std::string(homedir), ".cabocharc");
      std::ifstream ifs(s.c_str());
      if (ifs) return s;
    }

    char *rcenv = getenv ("CABOCHARC");
    if (rcenv) return std::string(rcenv);
#endif

#if defined  (_WIN32) && ! defined (__CYGWIN__)
    std::string r = getWindowsRegValue ("software\\cabocha", "cabocharc");
    if (! r.empty()) return r;
#endif

    return std::string (CABOCHA_DEFAULT_RC);
  }

  template <class T> static Analyzer *newAnalyzer (Param &p)
  {
    T *a = new T;
    if (! a->open (p)) throw std::runtime_error (a->what ());
    return (Analyzer *)a;
  }

#define _CABOCHA_INIT_PARSER tree(0), outputFormat(0), inputLayer(0), outputLayer(0)

#define PARSER_ERROR  std::ostrstream os; \
                      os << "Parser::open(): " << param.what () << "\n\n" \
                         <<  COPYRIGHT << "\ntry '--help' for more information.\n" << std::ends;  \
                      _what = os.str(); os.freeze (false);

  Parser::Parser (): _CABOCHA_INIT_PARSER {};

  Parser::Parser (int argc, char **argv): _CABOCHA_INIT_PARSER
  {
    if (! open (argc, argv)) throw std::runtime_error (_what);
  }

  bool Parser::open (int argc, char **argv)
  {
    Param param;

    if (! param.open (argc, argv, long_options)) {
      PARSER_ERROR;
      return false;
    }

    return open (param);
  }

  bool Parser::open (const char *arg)
  {
    Param param;

    if (! param.open (arg, long_options)) {
      PARSER_ERROR;
      return false;
    }

    return open (param);
  }

  Parser::~Parser()
  {
    close ();
  }

#define REPLACE_PROFILE(p,r,k) { \
   std::string tmp = (p).getProfileString (k); \
   replaceString (tmp, "$(rcpath)", r); \
   (p).setProfile (k, tmp.c_str(), true); }

  bool Parser::open (Param &param)
  {
    try {

      if (param.getProfileInt ("help")) {
        std::ostrstream ostrs;
        param.help (ostrs, long_options);
	ostrs << std::ends;
        std::runtime_error e (ostrs.str());
	ostrs.freeze (false);
	throw e;
      }

      if (param.getProfileInt ("version")) {
        std::ostrstream ostrs;
        param.version (ostrs, long_options);
	ostrs << std::ends;
        std::runtime_error e (ostrs.str());
	ostrs.freeze (false);
	throw e;
      }

      std::string rcfile = param.getProfileString ("rcfile");
      if (rcfile.empty()) rcfile = getDefaultRc ();
      if (! param.load (rcfile.c_str())) {
	PARSER_ERROR;
	return false;
      }

      std::string rcpath = rcfile;
      removeFileName (rcpath);

      REPLACE_PROFILE (param,rcpath,"parser-model");
      REPLACE_PROFILE (param,rcpath,"chunker-model");
      REPLACE_PROFILE (param,rcpath,"ne-model");
      REPLACE_PROFILE (param,rcpath,"chasenrc");

      close ();

      inputLayer   = param.getProfileInt ("input-layer");
      outputLayer  = param.getProfileInt ("output-layer");
      outputFormat = param.getProfileInt ("output-format");
      int action   = param.getProfileInt ("action-mode");
      int ne       = param.getProfileInt ("ne");

      if (action == SELECTION_MODE) outputFormat = FORMAT_NONE;

      std::string morphological_analyzer = param.getProfileString ("morphological-analyzer");

      int mmode = 0;
      if (morphological_analyzer == "chasen")     mmode = 0;
      else if (morphological_analyzer == "mecab") mmode = 1;
      else if (morphological_analyzer.empty())    mmode = 0;
      else {
	std::string e = "unknown morphological analyzer: " ; e += morphological_analyzer;
	throw std::runtime_error (e);
      }

      switch (inputLayer) {

      case INPUT_RAW_SENTENCE:  // case 1
	{
	  switch (outputLayer) {
	  case OUTPUT_POS:
	    analyzer.push_back (mmode == 0 ? newAnalyzer<ChaSen>(param) : newAnalyzer<MeCab>(param));
	    if (ne) analyzer.push_back (newAnalyzer<NE>(param));
	    break;
	  case OUTPUT_CHUNK:
	    analyzer.push_back (mmode == 0 ? newAnalyzer<ChaSen>(param) : newAnalyzer<MeCab>(param));
	    if (ne) analyzer.push_back (newAnalyzer<NE>(param));
	    analyzer.push_back (newAnalyzer<Chunker>(param));
	    break;
	  case OUTPUT_SELECTION:
	    analyzer.push_back (mmode == 0 ? newAnalyzer<ChaSen>(param) : newAnalyzer<MeCab>(param));
	    if (ne) analyzer.push_back (newAnalyzer<NE>(param));
	    analyzer.push_back (newAnalyzer<Chunker>(param));
	    analyzer.push_back (newAnalyzer<Selector>(param));
	    break;
	  case OUTPUT_PARSE:
	    analyzer.push_back (mmode == 0 ? newAnalyzer<ChaSen>(param) : newAnalyzer<MeCab>(param));
	    if (ne) analyzer.push_back (newAnalyzer<NE>(param));
	    analyzer.push_back (newAnalyzer<Chunker>(param));
	    analyzer.push_back (newAnalyzer<Selector>(param));
	    analyzer.push_back (newAnalyzer<DepAnalyzer>(param));
	    break;
	  }
	  break;
	}

      case INPUT_POS: // case 2
	{
	  if (ne) analyzer.push_back (newAnalyzer<NE>(param));
	  switch (outputLayer) {
	  case OUTPUT_POS:
	    break;
	  case OUTPUT_CHUNK:
	    analyzer.push_back (newAnalyzer<Chunker>(param));
	    break;
	  case OUTPUT_SELECTION:
	    analyzer.push_back (newAnalyzer<Chunker>(param));
	    analyzer.push_back (newAnalyzer<Selector>(param));
	    break;
	  case OUTPUT_PARSE:
	    analyzer.push_back (newAnalyzer<Chunker>(param));
	    analyzer.push_back (newAnalyzer<Selector>(param));
	    analyzer.push_back (newAnalyzer<DepAnalyzer>(param));
	    break;
	  }
	  break;
	}

      case INPUT_CHUNK: // case 3
	{
	  switch (outputLayer) {
	  case OUTPUT_POS:
	  case OUTPUT_CHUNK:
	    break;
	  case OUTPUT_SELECTION:
	    analyzer.push_back (newAnalyzer<Selector>(param));
	    break;
	  case OUTPUT_PARSE:
	    analyzer.push_back (newAnalyzer<Selector>(param));
	    analyzer.push_back (newAnalyzer<DepAnalyzer>(param));
	    break;
	  }
	  break;
	}

      case  INPUT_SELECTION: // case 4
	{
	  switch (outputLayer) {
	  case OUTPUT_POS:
	  case OUTPUT_CHUNK:
	  case OUTPUT_SELECTION:
	    break;
	  case OUTPUT_PARSE:
	    analyzer.push_back (newAnalyzer<DepAnalyzer>(param));
	    break;
	  }
	  break;
	}
      }

      tree = new Tree;

      return true;
    }

    catch (std::exception &e) {
      close ();
      _what = std::string ("Parser::open (): ") + e.what();
      return false;
    }
  }

  bool Parser::close ()
  {
    for (size_t i = 0; i < analyzer.size(); i++) delete analyzer[i];
    analyzer.clear();

    delete tree;
    tree = 0;

    outputFormat = 0;
    inputLayer = 0;
    outputLayer = 0;

    return true;
  }

  const char *Parser::what ()
  {
    return _what.c_str ();
  }

  int Parser::parse (int argc, char **argv)
  {
    try {

      Param param;

      if (! param.open (argc, argv, long_options)) {
        PARSER_ERROR;
        throw std::runtime_error (_what);
      }

      if (param.getProfileInt ("help")) {
        param.help (std::cout, long_options);
        return EXIT_SUCCESS;
      }

      if (param.getProfileInt ("version")) {
        param.version (std::cout, long_options);
        return EXIT_SUCCESS;
      }

      if (! open (param)) throw std::runtime_error (_what);

      //    std::cout.sync_with_stdio (false);
      std::ostream *ofs = &std::cout;

      std::string outputFileName = param.getProfileString ("output");

      if (! outputFileName.empty()) {
	ofs = new std::ofstream (outputFileName.c_str());
	if (!ofs) throw std::runtime_error (outputFileName + ", cannot open");
      }

      const std::vector <std::string>& rest = param.getRestArg ();

      if (rest.size()) {
	for (size_t i = 0; i < rest.size(); i++) {
	  std::ifstream ifs (rest[i].c_str ());
	  if (!ifs) throw std::runtime_error (rest[i] + ", cannot open");
	  while (int r = parse (ifs, *ofs)) {
	    if (r == -1) throw std::runtime_error (_what);
	  }
	}
      } else {
	while (int r = parse (std::cin, *ofs)) {
	  if (r == -1) throw std::runtime_error (_what);
	}
      }

      if (ofs != &std::cout) delete ofs;

      return EXIT_SUCCESS;
    }

    catch (std::exception &e) {
      std::cerr << "FATAL: " << e.what () << std::endl;
      return EXIT_FAILURE;
    }
  }

  // bare interface of parse
  Tree *Parser::parse (Tree *_tree)
  {
    if (!_tree) throw std::runtime_error ("Parser::parse(): NULL tree is given");
    for (size_t i = 0; i < analyzer.size(); i++) {
      analyzer[i]->parse (_tree);
    }
    return _tree;
  }

  Tree *Parser::parse (std::istream &is)
  {
    try {
      if (! tree->read (is, inputLayer)) {
	_what = std::string ("Parser::parse(): cannot read from input stream");
	return 0;
      }
      return parse (tree);
    }

    catch (std::exception &e) {
      _what = std::string ("Parser::parse(): ") + e.what();
      return 0;
    }
  }

  int Parser::parse (std::istream &is, std::ostream &os)
  {
    try {
      if (! tree->read(is, inputLayer)) return 0; // just retern 0
      parse (tree);
      tree->write (os, outputLayer, outputFormat); // safe to write
      return 1;
    }

    catch (std::exception &e) { // un-expected errors
      _what = std::string ("Parser::parse(): ") + e.what();
      return -1; // un-expected errors
    }
  }

  Tree* Parser::parse (const char *str, size_t len)
  {
    if (!str) {
      _what = "Parser::parse(): NULL string is given";
      return 0;
    }
    std::istrstream is (str, len ? len : strlen (str));
    return parse (is);
  }

  const char *Parser::parseToString (const char *str, size_t len, char *out, size_t len2)
  {
    if (!str) {
      _what = "Parser::parse (): NULL pointer is given";
      return 0;
    }

    std::istrstream is (str, len ? len : strlen (str));
    if (! parse (str, len)) return 0;
    std::ostrstream os (out, len2);
    tree->write (os); // safe to write
    os << std::ends;

    if (os.eof ()) {
      _what = "Parser::parse (): output buffer overflow" ;
      return 0;
    }

    return const_cast<char *>(out);
  }

  const char *Parser::parseToString (const char* str, size_t len)
  {
    if (! parse (str, len)) return 0;
    return tree->toString (outputFormat);
  }
}
