/*
  CaboCha -- Yet Another Japanese Dependency Structure Analyzer

  $Id: tree.cpp,v 1.15 2005/12/24 12:30:35 taku Exp $;

  Copyright (C) 2000-2004 Taku Kudo <taku-ku@is.aist-nara.ac.jp>
  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <cabocha.h>
#include <common.h>
#include <stdexcept>
#include <algorithm>
#include <vector>
#include <freelist.h>

#define CABOCHA_TREE_CHAR_MEM_SIZE  524288
#define CABOCHA_TREE_CHUNK_MEM_SIZE 256
#define CABOCHA_TREE_LINE_SIZE      1024

namespace CaboCha
{
  Tree::Tree (): freelist1(0), freelist2(0), ostrs (0)
  {
    freelist1 = new FreeList<char>  (CABOCHA_TREE_CHAR_MEM_SIZE);
    freelist2 = new FreeList<Chunk> (CABOCHA_TREE_CHUNK_MEM_SIZE);
    reserve (256);
  }

  Tree::~Tree ()
  {
    delete freelist1;
    delete freelist2;
    if (ostrs) { ostrs->freeze(false); delete ostrs; }
  }

  void Tree::clear()
  {
    //    std::vector <Token>::clear();
    this->erase(begin(), end());
    freelist1->free ();
    freelist2->free ();
  }

  char *Tree::alloc (size_t size)
  {
    return freelist1->alloc (size + 1);
  }

  char *Tree::alloc (char *str)
  {
    size_t size = strlen (str);
    char *n = freelist1->alloc (size + 1);
    strcpy (n, str);
    return n;
  }

  Chunk *Tree::allocChunk ()
  {
    Chunk *c = freelist2->alloc (1);
    c->link = -1;
    c->head = c->func = 0;
    c->score = 0.0;
    c->rel = this->alloc ("O");
    c->feature[0] = 0;
    return c;
  };

  ////////////////////////////////////////////////
  //
  // Input Lattice Format
  //
  std::istream &Tree::read  (std::istream &is, int inputLayer)
  {
    try {

      switch (inputLayer) {

	// input RAW sentence
      case INPUT_RAW_SENTENCE:
	if (! std::getline(is, sentence)) {
	  is.clear(std::ios::eofbit|std::ios::badbit);
	  return is;
	}
	break;

	// input POS/Chunk/Selector/Parser
      case INPUT_POS:
      case INPUT_CHUNK:
      case INPUT_SELECTION:
      case INPUT_PARSE:

	int next_status = TOKEN_ANY;
	Chunk *chunk    = 0;
	size_t ti = 0;
	char *column[64];

	clear ();

	// process each line
	for (;;) {

	  char *line = this->alloc (CABOCHA_TREE_LINE_SIZE);

	  // read one line
	  if (! is.getline (line, CABOCHA_TREE_LINE_SIZE)) {
	    is.clear(std::ios::eofbit|std::ios::badbit);
	    return is;
	  }

	  // is end of morphs?
	  if (line[0] == '\0'  || ! strcmp (line, "EOS")) return is;

	  // spllit
	  size_t _size = tokenize (line, "\t ", column, 64);

	  // chunk bonudary
	  if (_size >= 3 && ! strcmp (column[0], "*") && (column[1][0] == '-' || isdigit (column[1][0])) ) {

	    int id = str2int (column[1]);

	    if (id == -1) {
	      next_status = TOKEN_NEVER;
	      continue;
	    }

	    chunk = this->allocChunk ();

	    // link
	    chunk->link = str2int (column[2]);

	    // rel
	    size_t i;
	    size_t len = strlen (column[2]);
	    for (i = 0; i < len; i++)
	      if (! ((column[2][i] >= '0' && column[2][i] <= '9') || column[2][i] == '+' || column[2][i] == '-')) break;
	    chunk->rel = column[2] + i;

	    // head/func
	    if (_size >= 4) {
	      size_t i;
	      size_t len = strlen (column[3])-1;
	      for (i = 0; i < len ; i++) if (column[3][i] == '/') break;
	      chunk->head = str2int (column[3]);
	      chunk->func = str2int (column[3]+i+1);
	    }

	    // score;
	    if (_size == 5) chunk->score = atof (column[4]);

	    // feature, must set more than 2 features (ad-hoc)
	    if (_size >= 6) {
	      std::copy (column + 4, column + _size, chunk->feature);
	      chunk->feature[_size-4] = 0; // NULL terminate
	    }

	  } else if (_size >= 4) { // lists of morph

	    resize (ti + 1);

	    (*this)[ti].chunk    = chunk;
	    (*this)[ti].status   = next_status;
	    (*this)[ti].surface  = column[0];
	    (*this)[ti].read     = column[1];
	    (*this)[ti].base     = column[2];
	    (*this)[ti].pos      = column[3];
	    (*this)[ti].ctype    = column[4];
	    (*this)[ti].cform    = column[5];
	    (*this)[ti].ne       = column[6];
	    (*this)[ti].optional = column[7];

	    char *tmp = this->alloc (column[3]);
	    size_t s = tokenize (tmp, "-", (*this)[ti].pos_list, 8);
	    (*this)[ti].pos_list[s] = 0;

	    chunk = 0;
	    next_status = TOKEN_ANY;

	    ti++;

	    // other format, error
	  } else {
	    throw std::runtime_error (std::string ("format error: ") + std::string(line));
	  }
	}

	break;
      }

      return is;
    }

    catch (std::exception &e) {
      std::string what = std::string ("Tree::read(): ") + e.what();
      throw std::runtime_error (what);
      return is;
    }
  }

  const char *Tree::toString (int outputFormat)
  {
    if (! ostrs) ostrs = new std::ostrstream ();
    else { ostrs->freeze (false); ostrs->seekp (0, ios::beg); }
    this->write (*ostrs, this->outputLayer, outputFormat); // safe to write
    *ostrs << std::ends;
    return ostrs->str ();
  }

  const char *Tree::toString (int outputFormat, char *out, size_t len2)
  {
    std::ostrstream os (out, len2);
    this->write (os, this->outputLayer, outputFormat); // safe to write
    os << std::ends;
    return out;
  }

  std::ostream &Tree::write (std::ostream &os, int outputLayer, int outputFormat)
  {
    if (!os) throw std::runtime_error ("Tree::write(): cannot write to outputstream");

    switch (outputFormat) {
    case FORMAT_LATTICE:
      return writeLattice (os, outputLayer);
    case FORMAT_TREE_LATTICE:
      writeTree (os, outputLayer);
      return writeLattice (os, outputLayer);
    case FORMAT_TREE:
      return writeTree (os, outputLayer);
    case FORMAT_XML:
      return writeXML (os, outputLayer);
    case FORMAT_NONE:
      return os; // do nothing
    default:
      throw std::runtime_error ("Tree::write(): unkown format");
    }

    return os;
  }

  //////////////////////////////////////////////
  //
  // Output in Lattice Format
  //
  // Helper function for writeLattice
  //
  void inline static writeLatticeToken (Token &token, std::ostream &os)
  {
    os << token.surface << '\t' << token.read
       << '\t' << token.base << '\t' << token.pos << '\t'
       << token.ctype << '\t' << token.cform;
    if (token.ne[0]) os << '\t' << token.ne;
    os << '\n';
  }

  std::ostream &Tree::writeLattice (std::ostream &os, int outputLayer)
  {
    if (outputLayer == OUTPUT_RAW_SENTENCE) { // use such case?
      if (empty()) os << sentence << std::endl;
      else {
	for (size_t i = 0; i < size(); ++i) os << (*this)[i].surface;
	os << std::endl;
      }
      return os;
    }

    size_t ci = 0;

    os.setf(std::ios::fixed, std::ios::floatfield);
    os.precision(8);

    for (size_t i = 0; i < size(); i++) {
      if ((*this)[i].chunk && outputLayer != OUTPUT_POS) {
	Chunk *chunk = (*this)[i].chunk;
	os << "* " << ci++ << ' ' << chunk->link << chunk->rel; // basic

	switch (outputLayer) {
	case OUTPUT_SELECTION:
	  os << ' ' << chunk->head << "/" << chunk->func;
	  for (char **n = chunk->feature; *n; ++n) os << ' ' << *n;
	  break;
	case OUTPUT_PARSE:
	  os << ' ' << chunk->head << "/" << chunk->func << ' ' << chunk->score;
	  break;
	}

	os << '\n';
      }
      writeLatticeToken((*this)[i], os);
    }

    os << "EOS" << std::endl;
    return os;
  }

  void inline static writeXMLToken (Token &token, std::ostream &os, int i)
  {
    os << "  <tok id=\"" << i
       << "\" read=\"" << token.read
       << "\" base=\"" << token.base
       << "\" pos=\""  << token.pos << "\""
       << " ctype=\""  << token.ctype << "\""
       << " cform=\""  << token.cform << "\""
       << " ne=\""     << token.ne << "\""
       << ">" << token.surface << "</tok>\n";
  }

  std::ostream &Tree::writeXML (std::ostream &os, int outputLayer)
  {
    size_t ci = 0;

    os << "<sentence>" << std::endl;
    for (size_t i = 0; i < size(); ++i) {
      if ((*this)[i].chunk  && outputLayer != OUTPUT_POS) {
	if (ci) os << " </chunk>\n";

	Chunk *chunk = (*this)[i].chunk;
	os << " <chunk id=\"" << ci++ << "\" link=\"" << chunk->link
	   << "\" rel=\""   << chunk->rel
	   << "\" score=\"" << chunk->score
	   << "\" head=\""  << chunk->head + i
	   << "\" func=\""  << chunk->func + i << "\">\n";
      }
      writeXMLToken((*this)[i], os, i);
    }

    if (ci) os << " </chunk>\n";
    os << "</sentence>" << std::endl;
    return os;
  }

  //////////////////////////////////////////////
  //
  // Output in Tree Format
  //
  std::ostream &Tree::writeTree (std::ostream &os, int outputLayer)
  {
    if ((outputLayer != OUTPUT_PARSE) || (! empty() && ! (*this)[0].chunk))
      return writeLattice (os, outputLayer);

    // get max surface
    size_t max_len = 0;
    std::vector <std::pair <size_t, std::string> > chunks;

    size_t in = false;
    std::string ne = "";

    for (size_t i = 0; i < size();) {
      size_t cid = i;
      chunks.resize (chunks.size()+1);
      std::string surface = "";

      for (; i < size(); ++i) {

	if (in && ((*this)[i].ne[0] == 'B' || (*this)[i].ne[0] == 'O')) {
	  surface += "</";
	  surface += ne;
	  surface += ">";
	  in = false;
	}

	if (i != cid && (*this)[i].chunk) break;

	if ((*this)[i].ne[0] == 'B') {
	  ne = std::string ((*this)[i].ne + 2);
	  surface += "<";
	  surface += ne;
	  surface += ">";
	  in = true;
	}

	surface += std::string ((*this)[i].surface);

	if (in && i + 1 == size()) {
	  surface += "</";
	  surface += ne;
	  surface += ">";
	}
      }

      max_len = std::max ((size_t)max_len, surface.size());
      chunks[chunks.size()-1] = std::make_pair<size_t, std::string>(cid, surface);
    }

    std::vector <int> e(chunks.size());
    std::fill (e.begin(), e.end(), 0);

    for (size_t i = 0; i < chunks.size(); ++i) {

      const char *rel = (*this)[chunks[i].first].chunk->rel;
      int link        = (*this)[chunks[i].first].chunk->link;
      const char *surface = chunks[i].second.c_str();

      bool isdep = false;
      os.width(max_len + i*2);
      os.setf(std::ios::right,std::ios::adjustfield);
      os << surface;

      for (size_t j = i+1; j < chunks.size(); j++) {
        if (link == (int)j) {
          os << "-" << rel;
          isdep = 1;
          e[j] = 1;
        } else if (e[j]) {
          os <<  " |";
        } else if (isdep) {
          os << "  ";
        } else {
          os << "--";
        }
      }
      os << "\n";
    }

    os << "EOS" << std::endl;
    os.width (0);
    return os;
  }
};
